' Growing lists for Visual Basic.  Version 2.2, by Neil Fraser, October 96.
' The zeroth element of each list is the number of items in the list.
' The 'caseMatters' global variable (if any) determines case-sensitivity.

' This SUB adds the NewObject to the end of Array.
' This is the fastest method of adding.
' Eg. {"The", "", 7}, "Q" => {"The", "", 7, "Q"}
'
Sub ListAdd (Array(), NewObject)
  ListAddAt Array(), NewObject, Array(0) + 1
End Sub

' This SUB adds the NewObject string to Array at
' the requested Position.  Adding at position 1
' inserts at the front of the list, adding at
' position Array(0)+1 inserts at the end.
' Eg. {"The", "", 7}, "Q", 2 => {"The", "Q", "", 7}
'
Sub ListAddAt (Array(), NewObject, Position%)
  size% = Array(0) + 1
  If Position < 1 Or Position% > size% Then Stop ' Invalid inputs.
  ReDim Preserve Array(0 To size%) As Variant
  Array(0) = size%          ' Update the element counter
  
  For index% = size% - 1 To Position% Step -1' Shift any trailing elements away.
    Array(index% + 1) = Array(index%)
  Next index%
  Array(Position%) = NewObject   ' Add the NewObject
End Sub

' Copys the SourceArray to the TargetArray.
' Results in two identical lists.
'
Sub ListCopy (SourceArray(), TargetArray())
  size% = SourceArray(0)
  ReDim TargetArray(0 To size%) As Variant
  For index% = 0 To size%
    TargetArray(index%) = SourceArray(index%)
  Next index%
End Sub

' Separates a string of Words appart, and adds them individually to
' the Array.  In the proccess, it strips out all spaces.
' Eg. " A big,  fat 0" => {"A", "big,", "fat", "0"}
' Same as ListExplodeOn(words$, Array(), " "), but faster.
'
Sub ListExplode (words$, Array())
  words$ = Trim$(words$)
  div% = InStr(words$, " ")
  Do Until div% = 0
    ListAdd Array(), Left$(words$, div% - 1)
    words$ = LTrim$(Right$(words$, Len(words$) - div%))
    div% = InStr(words$, " ")
  Loop
  If words$ <> "" Then ListAdd Array(), words$
End Sub

' Separates a string of Words appart (see ListExplode),
' using the characters in sepatators$ to break the string.
' To break using whitespace, use this separator$:
' " "+Chr$(13)+Chr$(10)+Chr$(9)
'
Sub ListExplodeOn (words$, Array(), separators$)
  word$ = ""
  For index% = 1 To Len(words$)
    letter$ = Mid$(words$, index%, 1)
    If InStr(separators$, letter$) Then
      If word$ <> "" Then
        ListAdd Array(), word$
        word$ = ""
      End If
    Else
      word$ = word$ + letter$
    End If
  Next index%
  If word$ <> "" Then ListAdd Array(), word$
End Sub

' Iterates through the array and returns a string made up of
' the combined elements of the list.
' Eg. ListImplode$({"cat", "dog", "bird"}, ", ") -> "cat, dog, bird"
'
Function ListImplode$ (Array(), separator$)
  words$ = ""
  For index% = 1 To Array(0)
    words$ = words$ + separator$ + Array(index%)
  Next index%
  ' Oops, remove initial separator.
  If words$ <> "" Then ListImplode$ = Right$(words$, Len(words$) - Len(separator$))
End Function

' Loads the Array with data from a file.
' Opposite of ListSave Sub.
'
Sub ListLoad (Array(), Handle%)
  Input #Handle%, size%
  ReDim Array(0 To size%) As Variant
  Array(0) = size%
  For index% = 1 To size%
    Input #Handle%, Array(index%)
  Next index%
End Sub

' Return the position of the first Object in Array.
' Return zero if Object is not found.  The 'caseMatters'
' global variable (if any) determines case-sensitivity.
' Eg. {"A", "B", "C", "B"}, "B" => 2
'
Function ListPosition (Array(), Object)
  If Not caseMatters% Then Target$ = UCase$(Object)
  size% = Array(0)
  For index% = 1 To size%
    If caseMatters% Then
      If Array(index%) = Object Then ListPosition = index%: Exit Function
    Else
      If UCase$(Array(index%)) = Target$ Then ListPosition = index%: Exit Function
    End If
  Next index%
End Function

' Removes first instance of Victim from Array().
' Returns the former index of Victim, or zero if none existed.
' The 'caseMatters' global variable (if any) determines case-sensitivity.
'
Function ListRemoveItem (Array(), Victim)
  Target% = ListPosition(Array(), Victim) ' Find the victim
  If Target% Then ListRemoveLines Array(), Target%, Target%
  ListRemoveItem = Target%
End Function

' This Sub removes one line from Array.
'
Sub ListRemoveLine (Array(), Line1%)
  ListRemoveLines Array(), Line1%, Line1%
End Sub

' This Sub removes all lines from Line1 to Line2
' (inclusive).  If Line1 and Line2 are identicle, then
' only one line is removed.  The Array is redimed as
' needed.
'
Sub ListRemoveLines (Array(), Line1%, Line2%)
  If Line2 < Line1 Then ' What happened to the SWAP command?
    temp% = Line2%: Line2% = Line1%: Line1% = temp%
  End If
  size% = Array(0)
  If Line1% < 1 Or Line2% > size% Then Stop ' Invalid inputs
  
  deadLines% = Line2% - Line1% + 1
  For index% = Line2% + 1 To size%   ' Cascade tail forward
    Array(index% - deadLines%) = Array(index%)
  Next index%

  size% = size% - deadLines%
    ReDim Preserve Array(0 To size%) As Variant
  Array(0) = size%    ' Update the element counter
End Sub

' Reverse the order of the elements in the Array.
' Eg. {"C", "A", "B"} => {"B", "A", "C"}
'
Sub ListReverse (Array())
  ReDim temp(0) As Variant
  ListCopy Array(), temp()
  size% = Array(0)
  For index% = 1 To size%
    Array(index%) = temp(1 + size% - index%)
  Next index%
End Sub

' Return the position of the last Object in Array (Reverse Position).
' Return zero if Object is not found.  The 'caseMatters'
' global variable (if any) determines case-sensitivity.
' Eg. {"A", "B", "C", "B"}, "B" => 4
'
Function ListRPosition (Array(), Object)
  If Not caseMatters% Then Target$ = UCase$(Object)
  For index% = Array(0) To 1 Step -1
    If caseMatters% Then
      If Array(index%) = Object Then ListRPosition = index%: Exit Function
    Else
      If UCase$(Array(index%)) = Target$ Then ListRPosition = index%: Exit Function
    End If
  Next index%
End Function

' Saves the data in Array to a file.
' Opposite of ListLoad Sub.
'
Sub ListSave (Array(), Handle%)
  For index% = 0 To Array(0)
    Print #Handle%, Array(index%)
  Next index%
End Sub

' This Sub checks to see if the NewObject is already in Array.
' If it is, nothing happens, otherwise it is added to the end of Array$.
' The 'caseMatters' global variable (if any) determines case-sensitivity.
'
Sub ListSetAdd (Array(), NewObject)
  If (ListPosition(Array(), NewObject) = 0) Then
    ListAddAt Array(), NewObject, (Array(0) + 1)
  End If
End Sub

' Sorts the array according to the mode$:
'   ASCII   - alphabetic - low to hi
'   !ASCII  - alphabetic - hi to low
'   LENGTH  - number of chars - low to hi
'   !LENGTH - number of chars - hi to low
'   VAL     - numerical order - low to hi
'   !VAL    - numerical order - hi to low
' This sub uses a simple bubble sort.  Someone should upgrade it.
'
Sub ListSort (Array(), mode$)
  Do
    didSwap% = 0
    For index% = 2 To Array(0)
      doSwap% = ListSortEvaluate(Array(index% - 1), Array(index%), mode$)
      If doSwap% Then ' What happened to the SWAP command?
        temp = Array(index% - 1)
        Array(index% - 1) = Array(index%)
        Array(index%) = temp
        didSwap% = 1
      End If
    Next index%
  Loop Until didSwap% = 0
End Sub

' Sorts master array according to the mode$ (see ListSort for details).
' Mirrors the sort onto slave1 - used for sorting pairs of lists.
' This sub uses a simple bubble sort.  Someone should upgrade it.
'
Sub ListSortAnd (Master(), Slave1(), mode$)
  If Master(0) <= Slave1(0) Then
    Do
      didSwap% = 0
      For index% = 2 To Master(0)
        doSwap% = ListSortEvaluate(Master(index% - 1), Master(index%), mode$)
        If doSwap% Then ' What happened to the SWAP command?
          temp = Master(index% - 1)
          Master(index% - 1) = Master(index%)
          Master(index%) = temp
          temp = Slave1(index% - 1)
          Slave1(index% - 1) = Slave1(index%)
          Slave1(index%) = temp
          ' to create a ListSortAndAnd sub, just add an extra swap here.
          didSwap% = 1
        End If
      Next index%
    Loop Until didSwap% = 0
  End If
End Sub

' A private function used by the sorting subroutines.
' Returns 1 iff a and b are in the wrong order according to the mode$.
'
Function ListSortEvaluate (a, b, mode$)
  doSwap% = 0
  Select Case mode$
  Case "ASCII"
    If caseMatters% Then
      If a > b Then doSwap% = 1
    Else
      If UCase$(a) > UCase$(b) Then doSwap% = 1
    End If
  Case "!ASCII"
    If caseMatters% Then
      If a < b Then doSwap% = 1
    Else
      If UCase$(a) < UCase$(b) Then doSwap% = 1
    End If
  Case "LENGTH"
    If Len(a) > Len(b) Then doSwap% = 1
  Case "!LENGTH"
    If Len(a) < Len(b) Then doSwap% = 1
  Case "VAL"
    If Val(a) > Val(b) Then doSwap% = 1
  Case "!VAL"
    If Val(a) < Val(b) Then doSwap% = 1
  Case Else
    Stop ' Huh?
  End Select
  ListSortEvaluate = doSwap%
End Function

' Removes all duplicate strings from Array.
' The 'caseMatters' global variable (if any) determines case-sensitivity.
' Eg. {"cat", "dog", "cat"} => {"cat", "dog"}
'
Sub ListStripDups (Array())
  For index% = Array(0) To 2 Step -1
    If ListPosition(Array(), Array(index%)) <> index% Then
      ListRemoveLines Array(), index%, index%
    End If
  Next index%
End Sub

' Removes all duplicate strings from Array (see ListStripDups for details).
' Mirrors the operation onto slave1 - used for striping pairs of lists.
' Eg. {"c", "d", "c"}, {4, 2, 5} => {"c", "d"}, {4, 2}
'
Sub ListStripDupsAnd (Master(), Slave1())
  For index% = Master(0) To 2 Step -1
    If ListPosition(Master(), Master(index%)) <> index% Then
      ListRemoveLines Master(), index%, index%
      ListRemoveLines Slave1(), index%, index%
      ' to create a ListStripDupsAndAnd sub, just add an extra ListRemoveLines here.
    End If
  Next index%
End Sub

' Exchanges Array1 with Array2.
' Useful in an Undo command.
'
Sub ListSwap (Array1(), Array2())
  ReDim temp(0) As Variant
  ListCopy Array1(), temp()
  ListCopy Array2(), Array1()
  ListCopy temp(), Array2()
End Sub

